# -*- coding: utf-8 -*-
"""
Created on Thu May 13 11:59:29 2021

@author: bgregor
"""

# Count letter frequency in 1M words

 
# The multiprocessing library
import multiprocessing as mp
import string

# A function to download a file of 1M words
from io import BytesIO
from zipfile import ZipFile
from urllib.request import urlopen
import itertools as its
import pprint

from get_n_cores import get_n_cores

from rcs_timer import timer


def get_data():
    ''' Fetches a ZIP file, unpacks it to a file in the temp directory,
        returns the opened file. The data file is from SuperFastPython,
        an extensive tutorial on using multiprocessing:
            https://superfastpython.com/category/multiprocessing/
    '''
    resp = urlopen('https://raw.githubusercontent.com/SuperFastPython/DataSets/main/bin/1m_words.txt.zip')
    myzip = ZipFile(BytesIO(resp.read()))
    all_lines = [word.decode('utf-8') for word in myzip.open('1m_words.txt').readlines()]
    # Strip and lowercase the strings to remove newline characters
    return [word.strip().lower() for word in all_lines]


def count_letters(words):
    ''' Return a dictionary of every lowercase letter
    in English found in a list of words'''
    # hint: here's a built-in Python list of the English alphabet
    alphabet = list(string.ascii_lowercase)
    # and here's how to initialize a dictionary of {'a':0,'b':0,...}
    # using the itertools library without a 'for' loop. dict() will
    # create a dictionary from a list of (key,value) tuples
    letter_counts = dict(zip(alphabet,its.repeat(0)))
    # and fill it in...
    for word in words:
        for w in word:
            if w in letter_counts: # check in case w is punctuation
                letter_counts[w] += 1 
    return letter_counts    

@timer
def par_count_letters(words, ncores):
    ''' A parallel version of count_letters '''
    # Note - during development don't use the whole data set!  This is what
    # was causing my laptop to run out of memory...
    #     words = words[0:10000]
    # To split up the work with the pool we need a list of lists.
    # count_letters expects a list of words. It returns a dictionary.
    # If we modify count_letters to examine just a single word (instead of a
    # list of them) we'll get back 1M dictionaries that need to be joined. 
    # Too many!  How about splitting up into lists of 1000 words each.
    list_chunksize = 1000
    # You could also split up by ncores
    #    list_chunksize = len(words) // ncores
    # A one-line list comprehension does the trick.
    list_words = [words[i:i + list_chunksize] for i in range(0, len(words), list_chunksize)] 
    
    with mp.Pool(processes=ncores) as pool:
        list_counts = pool.map(count_letters, list_words)
    # list_counts is now a list of dictionaries, each with 26 letters as defined in 
    # count_letters. Aggregate them together into one dictionary.
    # Remove the last dictionary from the list, the others will get added
    # into it.
    result = list_counts.pop()
    for lc in list_counts:
        # lc is a dictionary. Loop over its keys, and use them to 
        # update result.
        for key in lc:
            result[key] += lc[key]
    
    return result 
    

@timer
def call_count_letters(words):
    ''' This is just to allow for a timer to be run
    on count_letters. If we add @timer to count_letters
    itself it'll print timing everytime the count_letters
    is called from the parallel pool'''
    return count_letters(words)

if __name__ == '__main__':
    # Fetch the data.
    words = get_data()
    # Print out the dictionary of letter counts:
    letter_counts = call_count_letters(words)
    pp = pprint.PrettyPrinter(indent=4)
    pp.pprint(letter_counts)    

    # Now do it again in parallel...
    ncores = get_n_cores()
    letter_counts_par = par_count_letters(words, ncores)
    # and print it - the results should be the same!
    pp.pprint(letter_counts_par)    
